/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2002 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */
package org.objectweb.medor.optim.jorm;

import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.Field;
import org.objectweb.medor.filter.api.Expression;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.filter.jorm.lib.CompositePName;
import org.objectweb.medor.filter.jorm.lib.SinglePName;
import org.objectweb.medor.filter.lib.And;
import org.objectweb.medor.filter.lib.BasicFieldOperand;
import org.objectweb.medor.filter.lib.BasicOperand;
import org.objectweb.medor.filter.lib.BasicParameterOperand;
import org.objectweb.medor.filter.lib.Equal;
import org.objectweb.medor.filter.lib.MemberOf;
import org.objectweb.medor.filter.lib.NavigatorOperator;
import org.objectweb.medor.optim.api.QueryRewriter;
import org.objectweb.medor.optim.lib.BasicQueryRewriter;
import org.objectweb.medor.optim.lib.FlattenQueryTreeRule;
import org.objectweb.medor.query.TestQueryTreeHelper;
import org.objectweb.medor.query.jorm.lib.NavigatorNodeFactory;
import org.objectweb.medor.query.jorm.lib.JormQueryTreeHelper;
import org.objectweb.medor.query.jorm.api.JormExtent;
import org.objectweb.medor.query.api.OrderField;
import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.lib.BasicOrderField;
import org.objectweb.medor.query.lib.QueryTreePrinter;
import org.objectweb.medor.query.lib.SelectProject;
import org.objectweb.medor.query.rdb.api.QualifiedTable;
import org.objectweb.medor.query.rdb.api.RdbExpField;
import org.objectweb.medor.query.rdb.api.RdbExpQueryLeaf;
import org.objectweb.medor.query.rdb.lib.BasicQualifiedTable;
import org.objectweb.medor.query.rdb.lib.BasicRdbExpQueryLeaf;
import org.objectweb.medor.type.lib.PTypeSpaceMedor;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.ArrayList;
import java.util.Iterator;

/**
 * @author S.Chassande-Barrioz
 */
public abstract class TestJormQueryRewriter extends TestJormPDHelper {

	public TestJormQueryRewriter(String testName, String loggerName) {
		super(testName, loggerName);
	}

	public abstract QueryLeaf getQueryLeafA(Example e) throws MedorException;

	public abstract QueryLeaf getQueryLeafOrderByA(Example e)
			throws MedorException;

	public abstract QueryLeaf getQueryLeafMemberOfBD(Example eB, Example eD)
			throws MedorException;

	public abstract QueryLeaf getQueryLeafMemberOfBC(Example eB, Example eC)
			throws MedorException;

	public void test1() {
		try {
			Example e = getExtentOfA();
			Expression exp = new Equal(
					new BasicFieldOperand(e.extent.getField(e.clazz.getFQName() + ".field1")),
					new BasicParameterOperand(PTypeSpace.STRING, "?1"));
			QueryNode qn = new SelectProject(exp, "", null);
			QueryTreeField qtf = (QueryTreeField)
					e.extent.getField(e.extent.getPNameFieldName());
			qn.addPropagatedField(
					qtf.getName(),
					qtf.getType(),
					new QueryTreeField[]{qtf});
			ArrayList rules = new ArrayList(2);
			rules.add(new FlattenQueryTreeRule());
			rules.add(new JormFlatten2Rdb());
			QueryRewriter jqr = new BasicQueryRewriter(rules);

			logger.log(BasicLevel.DEBUG, "test1: QueryTree before transformation");
			QueryTreePrinter.printQueryTree(qn, logger);
			//assign mapper + project names to each extent
			e.extent.setPMapper(mapper, PROJECT_NAME);
			e.extent.setProjectName("test");
			QueryTree result = jqr.transform(qn);
			logger.log(BasicLevel.DEBUG, "QueryTree rewritten");

			QueryTreePrinter.printQueryTree(result, logger);
			QueryLeaf leaf = getQueryLeafA(e);
			QueryNode expected = new SelectProject("", null);
			expected.addCalculatedField(
					e.clazz.getFQName() + ".object",
					e.extent.getField(e.extent.getPNameFieldName()).getType(),
					new SinglePName(
							new BasicFieldOperand(
									leaf.getTupleStructure().getField(e.clazz.getFQName()
					+ "." + "id")),
							new BasicParameterOperand(
									PTypeSpaceMedor.NAMING_CONTEXT,
									"A"),
							e.extent.getField(e.clazz.getFQName() + ".object").getType()
					)
			);
			logger.log(BasicLevel.DEBUG, "test1: Expected QueryTree");
			logger.log(BasicLevel.DEBUG, "=========================================");
			QueryTreePrinter.printQueryTree(expected, logger);
			TestQueryTreeHelper.equals("", expected, result, logger);
		} catch (MedorException e1) {
			e1.printStackTrace();
			fail(e1.getMessage());
		}
	}

	public void test1distinct() {
		try {
			Example e = getExtentOfA();
			Expression exp = new Equal(
					new BasicFieldOperand(e.extent.getField(e.clazz.getFQName() + ".field1")),
					new BasicParameterOperand(PTypeSpace.STRING, "?1"));
			QueryNode qn = new SelectProject(exp, "", null);
			qn.setDistinct(true);
			QueryTreeField qtf = (QueryTreeField)
					e.extent.getField(e.extent.getPNameFieldName());
			qn.addPropagatedField(
					qtf.getName(),
					qtf.getType(),
					new QueryTreeField[]{qtf});
			ArrayList rules = new ArrayList(2);
			rules.add(new FlattenQueryTreeRule());
			rules.add(new JormFlatten2Rdb());
			QueryRewriter jqr = new BasicQueryRewriter(rules);

			logger.log(BasicLevel.DEBUG, "test1distinct: QueryTree before transformation");
			QueryTreePrinter.printQueryTree(qn, logger);
			//assign mapper + project names to each extent
			e.extent.setPMapper(mapper, PROJECT_NAME);
			e.extent.setProjectName("test");
			QueryTree result = jqr.transform(qn);
			logger.log(BasicLevel.DEBUG, "QueryTree rewritten");

			QueryTreePrinter.printQueryTree(result, logger);
			QueryLeaf leaf = getQueryLeafA(e);
			QueryNode expected = new SelectProject("", null);
			expected.setDistinct(true);
			leaf.setDistinct(true);
			expected.addCalculatedField(
					e.clazz.getFQName() + ".object",
					e.extent.getField(e.extent.getPNameFieldName()).getType(),
					new SinglePName(
							new BasicFieldOperand(
									leaf.getTupleStructure().getField(e.clazz.getFQName()
					+ "." + "id")),
							new BasicParameterOperand(
									PTypeSpaceMedor.NAMING_CONTEXT,
									"A"),
							e.extent.getField(e.clazz.getFQName() + ".object").getType()
					)
			);
			logger.log(BasicLevel.DEBUG, "test1distinct: Expected QueryTree");
			logger.log(BasicLevel.DEBUG, "=========================================");
			QueryTreePrinter.printQueryTree(expected, logger);
			TestQueryTreeHelper.equals("", expected, result, logger);
		} catch (MedorException e1) {
			e1.printStackTrace();
			fail(e1.getMessage());
		}
	}

	public void testOrderBy() {
		try {
			Example e = getExtentOfA();
			Expression exp = new Equal(
					new BasicFieldOperand(e.extent.getField(e.clazz.getFQName() + ".field1")),
					new BasicParameterOperand(PTypeSpace.STRING, "?1"));
			QueryNode qn = new SelectProject(exp, "", null);
			QueryTreeField qtf = (QueryTreeField)
					e.extent.getField(e.extent.getPNameFieldName());
			qn.addPropagatedField(
					qtf.getName(),
					qtf.getType(),
					new QueryTreeField[]{qtf});
			QueryTreeField field1 =
					(QueryTreeField)e.extent.getField(e.clazz.getFQName() + ".field1");
			QueryTreeField field2 =
					(QueryTreeField)e.extent.getField(e.clazz.getFQName() + ".field2");
			// Add an OrderBy Field
			OrderField of1 = new BasicOrderField(field1, true);
			OrderField of2 = new BasicOrderField(field2);
			qn.setOrderBy(new OrderField[] {of1, of2});
			ArrayList rules = new ArrayList(2);
			rules.add(new FlattenQueryTreeRule());
			rules.add(new JormFlatten2Rdb());
			QueryRewriter jqr = new BasicQueryRewriter(rules);

			logger.log(BasicLevel.DEBUG, "test1distinct: QueryTree before transformation");
			QueryTreePrinter.printQueryTree(qn, logger);
			//assign mapper + project names to each extent
			e.extent.setPMapper(mapper, PROJECT_NAME);
			e.extent.setProjectName("test");
			QueryTree result = jqr.transform(qn);
			logger.log(BasicLevel.DEBUG, "QueryTree rewritten");

			QueryTreePrinter.printQueryTree(result, logger);
			QueryLeaf leaf = getQueryLeafOrderByA(e);
			QueryNode expected = new SelectProject("", null);
			expected.addCalculatedField(
					e.clazz.getFQName() + ".object",
					e.extent.getField(e.extent.getPNameFieldName()).getType(),
					new SinglePName(
							new BasicFieldOperand(
									leaf.getTupleStructure().getField(e.clazz.getFQName()
					+ "." + "id")),
							new BasicParameterOperand(
									PTypeSpaceMedor.NAMING_CONTEXT,
									"A"),
							e.extent.getField(e.clazz.getFQName() + ".object").getType()
					)
			);
			logger.log(BasicLevel.DEBUG, "test1distinct: Expected QueryTree");
			logger.log(BasicLevel.DEBUG, "=========================================");
			QueryTreePrinter.printQueryTree(expected, logger);
			TestQueryTreeHelper.equals("", expected, result, logger);
		} catch (MedorException e1) {
			e1.printStackTrace();
			fail(e1.getMessage());
		}
	}

	public void testMemberOfsingleBC() {
		try {
			Example eB = getExtentOfB(new String[] {"y"});
			Example eC = getExtentOfC();
			ArrayList al1 = new ArrayList(1);
			al1.add(new BasicFieldOperand(eB.extent.getField(eB.clazz.getFQName() + ".y")));
			ArrayList al2 = new ArrayList(1);
			al2.add(new BasicFieldOperand(eC.extent.getField(eC.clazz.getFQName() + ".x")));
			Expression exp1 = new MemberOf(al1, al2);
			Expression exp = new And(exp1,
					new Equal(new BasicFieldOperand(eC.extent.getField(eC.clazz.getFQName() + ".x")),
							new BasicOperand((long) 2)));
			QueryNode qn = new SelectProject(exp, "", null);
			QueryTreeField qtf = (QueryTreeField)
					eB.extent.getField(eB.extent.getPNameFieldName());
			qn.addPropagatedField(
					qtf.getName(),
					qtf.getType(),
					new QueryTreeField[]{qtf});
			ArrayList rules = new ArrayList(2);
			rules.add(new FlattenQueryTreeRule());
			rules.add(new JormFlatten2Rdb());
			QueryRewriter jqr = new BasicQueryRewriter(rules);

			logger.log(BasicLevel.DEBUG, "testMemberOfSingleBC: QueryTree before transformation");
			QueryTreePrinter.printQueryTree(qn, logger);
			//assign mapper + project names to each extent
			eB.extent.setPMapper(mapper, PROJECT_NAME);
			eB.extent.setProjectName("test");
			eC.extent.setPMapper(mapper, PROJECT_NAME);
			eC.extent.setProjectName("test");
			QueryTree result = jqr.transform(qn);
			logger.log(BasicLevel.DEBUG, "QueryTree rewritten");
			QueryTreePrinter.printQueryTree(result, logger);
			//TODO update expected QueryTree
			QueryLeaf leaf = getQueryLeafMemberOfBC(eB, eC);
			QueryNode expected = new SelectProject("", null);
			expected.addCalculatedField(
					eB.clazz.getFQName() + ".object",
					eB.extent.getField(eB.extent.getPNameFieldName()).getType(),
					new CompositePName(
							new FieldOperand[]{
								new BasicFieldOperand(
										leaf.getTupleStructure().getField(eB.clazz.getFQName()
										+ "." + "c2")),
								new BasicFieldOperand(
										leaf.getTupleStructure().getField(eB.clazz.getFQName()
										+ "." + "c1"))
							},
							new String[]{"c2","c1"},
							null,
							eB.extent.getField(eB.clazz.getFQName() + ".object").getType()
					)
			);
			logger.log(BasicLevel.DEBUG, "testMemberOfSingleBC: Expected QueryTree");
			logger.log(BasicLevel.DEBUG, "=========================================");
			QueryTreePrinter.printQueryTree(expected, logger);
			TestQueryTreeHelper.equals("", expected, result, logger);
		} catch (MedorException e1) {
			e1.printStackTrace();
			fail(e1.getMessage());
		}
	}

	public void testMemberOfmultiBD() {
		try {
			Example eB = getExtentOfB();
			Example eD = getExtentOfD();
			ArrayList al1 = new ArrayList(1);
			al1.add(new BasicFieldOperand(eB.extent.getField(eB.clazz.getFQName() + ".object")));
			ArrayList al2 = new ArrayList(1);
			al2.add(new BasicFieldOperand(eD.extent.getField(eD.clazz.getFQName() + ".object")));
			Expression exp = new MemberOf(al1, al2);
			QueryNode qn = new SelectProject(exp, "", null);
			QueryTreeField qtf = (QueryTreeField)
					eB.extent.getField(eB.extent.getPNameFieldName());
			qn.addPropagatedField(
					qtf.getName(),
					qtf.getType(),
					new QueryTreeField[]{qtf});
			ArrayList rules = new ArrayList(2);
			rules.add(new FlattenQueryTreeRule());
			rules.add(new JormFlatten2Rdb());
			QueryRewriter jqr = new BasicQueryRewriter(rules);

			logger.log(BasicLevel.DEBUG, "testMemberOf: QueryTree before transformation");
			QueryTreePrinter.printQueryTree(qn, logger);
			//assign mapper + project names to each extent
			eB.extent.setPMapper(mapper, PROJECT_NAME);
			eB.extent.setProjectName("test");
			eD.extent.setPMapper(mapper, PROJECT_NAME);
			eD.extent.setProjectName("test");
			QueryTree result = jqr.transform(qn);
			logger.log(BasicLevel.DEBUG, "QueryTree rewritten");
			QueryTreePrinter.printQueryTree(result, logger);

			//building the expected result
			QueryLeaf leaf = getQueryLeafMemberOfBD(eB, eD);
			QueryNode expected = new SelectProject("", null);
			expected.addCalculatedField(
					eB.clazz.getFQName() + ".object",
					eB.extent.getField(eB.extent.getPNameFieldName()).getType(),
					new CompositePName(
							new FieldOperand[]{
								new BasicFieldOperand(
										leaf.getTupleStructure().getField(eB.clazz.getFQName()
										+ "." + "c2")),
								new BasicFieldOperand(
										leaf.getTupleStructure().getField(eB.clazz.getFQName()
										+ "." + "c1"))
							},
							new String[]{"c2","c1"},
							null,
							eB.extent.getField(eB.clazz.getFQName() + ".object").getType()
					)
			);
			logger.log(BasicLevel.DEBUG, "testMemberOf: Expected QueryTree");
			logger.log(BasicLevel.DEBUG, "=========================================");
			QueryTreePrinter.printQueryTree(expected, logger);
			TestQueryTreeHelper.equals("", expected, result, logger);
		} catch (MedorException e1) {
			e1.printStackTrace();
			fail(e1.getMessage());
		}
	}

	public void testMemberOfParamBD() {
		try {
			logger.log(BasicLevel.DEBUG, "Entering testMemberOfParam");

			Example eB = getExtentOfB(new String[] {"y"});
			Example eD = getExtentOfD();
			ArrayList al1 = new ArrayList(1);
			BasicOperand value = new BasicOperand(1013);
			al1.add(value);
			ArrayList al2 = new ArrayList(1);
			al2.add(new BasicFieldOperand(eB.extent.getField(eB.clazz.getFQName() + ".y")));
			Expression exp = new MemberOf(al1, al2);
			QueryNode qn = new SelectProject(exp, "", null);
			QueryTreeField qtf = (QueryTreeField)
					eD.extent.getField(eD.extent.getPNameFieldName());
			qn.addPropagatedField(
					qtf.getName(),
					qtf.getType(),
					new QueryTreeField[]{qtf});
			ArrayList rules = new ArrayList(2);
			rules.add(new FlattenQueryTreeRule());
			rules.add(new JormFlatten2Rdb());
			QueryRewriter jqr = new BasicQueryRewriter(rules);

			logger.log(BasicLevel.DEBUG, "testMemberOf: QueryTree before transformation");
			QueryTreePrinter.printQueryTree(qn, logger);
			//assign mapper + project names to each extent
			eB.extent.setPMapper(mapper, PROJECT_NAME);
			eB.extent.setProjectName("test");
			eD.extent.setPMapper(mapper, PROJECT_NAME);
			eD.extent.setProjectName("test");
			QueryTree result = jqr.transform(qn);
			logger.log(BasicLevel.DEBUG, "QueryTree rewritten");
			QueryTreePrinter.printQueryTree(result, logger);

			//building the expected result
			QueryLeaf leaf = getQueryLeafMemberOfBD(eB, eD);
			QueryNode expected = new SelectProject("", null);
			expected.addCalculatedField(
					eB.clazz.getFQName() + ".object",
					eB.extent.getField(eB.extent.getPNameFieldName()).getType(),
					new CompositePName(
							new FieldOperand[]{
								new BasicFieldOperand(
										leaf.getTupleStructure().getField(eB.clazz.getFQName()
										+ "." + "c2")),
								new BasicFieldOperand(
										leaf.getTupleStructure().getField(eB.clazz.getFQName()
										+ "." + "c1"))
							},
							new String[]{"c2","c1"},
							null,
							eB.extent.getField(eB.clazz.getFQName() + ".object").getType()
					)
			);
			logger.log(BasicLevel.DEBUG, "testMemberOf: Expected QueryTree");
			logger.log(BasicLevel.DEBUG, "=========================================");
			QueryTreePrinter.printQueryTree(expected, logger);
			TestQueryTreeHelper.equals("", expected, result, logger);
		} catch (MedorException e1) {
			e1.printStackTrace();
			fail(e1.getMessage());
		}
	}

	public void testAtoB() {
		try {
			Example a = getExtentOfAcolocB();
			Example b1 = getExtentOfBcolocC();
			Example b2 = getExtentOfBcolocC();
			Expression exp1 = new Equal(
					new BasicFieldOperand(a.extent.getField(a.clazz.getFQName() + ".b1")),
					new BasicFieldOperand(b1.extent.getField(b1.extent.getPNameFieldName()))
			);
			Expression exp2 = new Equal(
					new BasicFieldOperand(a.extent.getField(a.clazz.getFQName() + ".b2")),
					new BasicFieldOperand(b2.extent.getField(b2.extent.getPNameFieldName()))
			);
			Expression exp3 = new And(exp1, exp2);
			Expression exp4 = new Equal(
					new BasicFieldOperand(b1.extent.getField(b1.clazz.getFQName() + ".f1")),
					new BasicFieldOperand(b2.extent.getField(b2.clazz.getFQName() + ".f2"))
			);
			Expression exp = new And(exp3, exp4);
			QueryNode qn = new SelectProject(exp, "", null);
			QueryTreeField f = (QueryTreeField) a.extent.getField(a.extent.getPNameFieldName());
			qn.addPropagatedField(f.getName(), f.getType(), new QueryTreeField[]{f});

			QueryRewriter jqr = this.createQueryRewriter();
			/*
			logger.log(BasicLevel.DEBUG, "testAtoB: QueryTree before transformation");
			logger.log(BasicLevel.DEBUG, "=========================================");
			QueryTreePrinter.printQueryTree(qn, logger);
			*/
			//assign mapper + project names to each extent
			a.extent.setPMapper(mapper, PROJECT_NAME);
			a.extent.setProjectName("test");
			b1.extent.setPMapper(mapper, PROJECT_NAME);
			b1.extent.setProjectName("test");
			b2.extent.setPMapper(mapper, PROJECT_NAME);
			b2.extent.setProjectName("test");
			QueryTree result = jqr.transform(qn);
			logger.log(BasicLevel.DEBUG, "QueryTree rewritten");
			QueryTreePrinter.printQueryTree(result, logger);
			logger.log(BasicLevel.DEBUG, "=========================================");
			Example ea = getExtentOfAcolocB();
			QualifiedTable tableA = new BasicQualifiedTable("T_A", "T_A_0");
			String nameA = "org.objectweb.medor.optim.jorm.rdb.AcolocB";
			String prefixA = nameA + ".";
			QualifiedTable tableB1 = new BasicQualifiedTable("T_B", "T_B_1_b1_join");
			QualifiedTable tableB2 = new BasicQualifiedTable("T_B", "T_B_2_b2_join");
			String nameB = "org.objectweb.medor.optim.jorm.rdb.BcolocC";
			String prefixB = nameB + ".";
			RdbExpQueryLeaf leaf = new BasicRdbExpQueryLeaf(null,
					new QualifiedTable[]{tableA, tableB1, tableB2}, "");
			RdbExpField idA = leaf.addRdbField(
					prefixA + "a_name", PTypeSpace.LONG, "A_PK", tableA);
			RdbExpField b1ToA = leaf.addRdbField(
					"BtoA_1_FK", PTypeSpace.LONG, "BtoA_1_FK", tableB1);
			RdbExpField f1 = leaf.addRdbField(
					prefixB + "f1", PTypeSpace.STRING, "col_f1", tableB1);
			leaf.removeRdbField(f1);
			RdbExpField b2ToA = leaf.addRdbField(
					"BtoA_2_FK", PTypeSpace.LONG, "BtoA_2_FK", tableB2);
			RdbExpField f2 = leaf.addRdbField(
					prefixB + "f2", PTypeSpace.STRING, "col_f2", tableB2);
			leaf.removeRdbField(f2);
			Expression eq3 = new Equal(
					new BasicFieldOperand(f1), new BasicFieldOperand(f2)
			);
			Expression eq1 = new Equal(
					new BasicFieldOperand(idA), new BasicFieldOperand(b1ToA)
			);
			Expression eq2 = new Equal(
					new BasicFieldOperand(idA), new BasicFieldOperand(b2ToA)
			);
			Expression expRdb = new And(new And(eq1, eq2), eq3);
			leaf.setQueryFilter(expRdb);
			QueryNode expected = new SelectProject("", null);
			SinglePName spn = new SinglePName(
					new BasicFieldOperand(idA),
					new BasicParameterOperand(
							PTypeSpaceMedor.NAMING_CONTEXT, prefixA + "object"),
					ea.extent.getField(prefixA + "object").getType());
			expected.addCalculatedField(prefixA + "object", spn.getType(), spn);

			logger.log(BasicLevel.DEBUG, "testAtoB: Expected QueryTree");
			logger.log(BasicLevel.DEBUG, "=========================================");

			QueryTreePrinter.printQueryTree(expected, logger);
			TestQueryTreeHelper.equals("", expected, result, logger);
			logger.log(BasicLevel.DEBUG, "End of testAtoB");

		} catch (MedorException e1) {
			e1.printStackTrace();
			fail(e1.getMessage());
		}
	}

	public void testAtoBtoC() {
		try {
			Example a = getExtentOfAcolocB();
			Example b = getExtentOfBcolocC();
			Example c = getExtentOfCcoloc();
			Expression exp1 = new Equal(
					new BasicFieldOperand(a.extent.getField(a.clazz.getFQName() + ".b1")),
					new BasicFieldOperand(b.extent.getField(b.extent.getPNameFieldName()))
			);
			Expression exp2 = new Equal(
					new BasicFieldOperand(b.extent.getField(b.clazz.getFQName() + ".c")),
					new BasicFieldOperand(c.extent.getField(c.extent.getPNameFieldName()))
			);
			Expression exp = new And(exp1, exp2);
			QueryNode qn = new SelectProject(exp, "", null);
			QueryTreeField f = (QueryTreeField) a.extent.getField(a.extent.getPNameFieldName());
			QueryTreeField fc = (QueryTreeField) c.extent.getField(c.clazz.getFQName() + ".fc");
			qn.addPropagatedField(f.getName(), f.getType(), new QueryTreeField[]{f});
			qn.addPropagatedField(fc.getName(), fc.getType(), new QueryTreeField[]{fc});

			ArrayList rules = new ArrayList(2);
			rules.add(new FlattenQueryTreeRule());
			rules.add(new JormFlatten2Rdb());
			QueryRewriter jqr = new BasicQueryRewriter(rules);
			logger.log(BasicLevel.DEBUG, "testBtoC: QueryTree before transformation");
			QueryTreePrinter.printQueryTree(qn, logger);
			//assign mapper + project names to each extent
			a.extent.setPMapper(mapper, PROJECT_NAME);
			a.extent.setProjectName("test");
			b.extent.setPMapper(mapper, PROJECT_NAME);
			b.extent.setProjectName("test");
			c.extent.setPMapper(mapper, PROJECT_NAME);
			c.extent.setProjectName("test");
			QueryTree result = jqr.transform(qn);
			logger.log(BasicLevel.DEBUG, "QueryTree rewritten");
			QueryTreePrinter.printQueryTree(result, logger);

			Example ea = getExtentOfAcolocB();
			QualifiedTable tableA = new BasicQualifiedTable("T_A", "T_A_0");
			String nameA = "org.objectweb.medor.optim.jorm.rdb.AcolocB";
			String prefixA = nameA + ".";
			QualifiedTable tableB = new BasicQualifiedTable("T_B", "T_B_1_b1_join");
			QualifiedTable tableC = new BasicQualifiedTable("T_C", "T_C_4_c_join");
			String nameB = "org.objectweb.medor.optim.jorm.rdb.BcolocC";
			String prefixB = nameB + ".";
			String nameC = "org.objectweb.medor.optim.jorm.rdb.Ccoloc";
			String prefixC = nameC + ".";
			RdbExpQueryLeaf leaf = new BasicRdbExpQueryLeaf(null,
					new QualifiedTable[]{tableA, tableC, tableB}, "");
			QueryTreeField idA = leaf.addRdbField(
					prefixA + "a_name", PTypeSpace.LONG, "A_PK", tableA);
			QueryTreeField b1ToA = leaf.addRdbField(
					"BtoA_1_FK", PTypeSpace.LONG, "BtoA_1_FK", tableB);
			QueryTreeField cToB = leaf.addRdbField(
					"CtoB_FK", PTypeSpace.STRING, "CtoB_FK", tableC);
			QueryTreeField idB = leaf.addRdbField(
					prefixB + "b_name", PTypeSpace.STRING, "B_PK", tableB);
			QueryTreeField fcField = leaf.addRdbField(
					prefixC + "fc", PTypeSpace.STRING, "FC_COL", tableC);
			Expression eq1 = new Equal(
					new BasicFieldOperand(idA), new BasicFieldOperand(b1ToA)
			);
			Expression eq2 = new Equal(
					new BasicFieldOperand(idB), new BasicFieldOperand(cToB)
			);
			Expression expRdb = new And(eq1, eq2);
			leaf.setQueryFilter(expRdb);
			QueryNode expected = new SelectProject("", null);
			SinglePName spn = new SinglePName(
					new BasicFieldOperand(idA),
					new BasicParameterOperand(
							PTypeSpaceMedor.NAMING_CONTEXT, prefixA + "object"),
					ea.extent.getField(prefixA + "object").getType());
			expected.addCalculatedField(prefixA + "object", spn.getType(), spn);
			expected.addPropagatedField(prefixC + "fc", fcField.getType(), new QueryTreeField[]{fcField});

			logger.log(BasicLevel.DEBUG, "testAtoB: Expected QueryTree");
			logger.log(BasicLevel.DEBUG, "=========================================");

			QueryTreePrinter.printQueryTree(expected, logger);
			TestQueryTreeHelper.equals("", expected, result, logger);
			logger.log(BasicLevel.DEBUG, "End of testAtoB");


		} catch (MedorException e1) {
			e1.printStackTrace();
			fail(e1.getMessage());
		}
	}

/*
public void _testCruise() {
		try {
			Example eRes = getExtentOfRes();
			Example eCruise = getExtentOfCruise();
			Example eCustomer = getExtentOfCustomer();

			Field f = eCruise.extent.getField(
				getFieldName(eCruise.extent.getName(), eCruise.pnameFieldName));
			NavigatorOperator nav1 = new NavigatorOperator(
					new BasicFieldOperand(f),
					eCruise.clazz.getFQName() + ".setOfRes");

			QueryTree qn1 = NavigatorNodeFactory.getIn(nav1, null, "cruise.resa");

			Field fResa = qn1.getTupleStructure().getField("cruise.resa");
			NavigatorOperator nav2 = new NavigatorOperator(
					new BasicFieldOperand(fResa),
					"setOfCust");
			QueryTree qn2 = NavigatorNodeFactory.getNavigatorNode(nav2, "Res", null);

			//now build the memberof  between the obtained customers and the
			//initial customer.
			ArrayList al1 = new ArrayList(1);
			al1.add(new BasicFieldOperand(eCustomer.extent.getField(eCustomer.clazz.getFQName() + ".object")));
			ArrayList al2 = new ArrayList(1);
			al2.add(new BasicFieldOperand(qn2.getTupleStructure().getField("Res.setOfCust.element")));
			Expression exp = new MemberOf(al1, al2);
			QueryNode qn = new SelectProject(exp, "", null);
			QueryTreeField qtf = (QueryTreeField)
					eCustomer.extent.getField(eCustomer.extent.getPNameFieldName());
			qn.addPropagatedField(
					qtf.getName(),
					qtf.getType(),
					new QueryTreeField[]{qtf});

			ArrayList rules = new ArrayList(2);
			rules.add(new FlattenQueryTreeRule());
			rules.add(new JormFlatten2Rdb());
			QueryRewriter jqr = new BasicQueryRewriter(rules);

			logger.log(BasicLevel.DEBUG, "testCruise: QueryTree before transformation");
			QueryTreePrinter.printQueryTree(qn, logger);

			//assign mapper + project names to each extent
			Iterator it = JormQueryTreeHelper.getJormExtents(qn).iterator();
			while (it.hasNext()) {
				JormExtent jext = (JormExtent) it.next();
				jext.setPMapper(mapper, PROJECT_NAME);
				jext.setProjectName("test");
			}
			eCruise.extent.setPMapper(mapper, PROJECT_NAME);
			eCruise.extent.setProjectName("test");
			eRes.extent.setPMapper(mapper, PROJECT_NAME);
			eRes.extent.setProjectName("test");
			eCustomer.extent.setPMapper(mapper, PROJECT_NAME);
			eCustomer.extent.setProjectName("test");

			QueryTree result = jqr.transform(qn);
			logger.log(BasicLevel.DEBUG, "QueryTree rewritten");
			QueryTreePrinter.printQueryTree(result, logger);

			//TODO build expected and compare

			logger.log(BasicLevel.DEBUG, "testCruise: Expected QueryTree");
			logger.log(BasicLevel.DEBUG, "=========================================");

			QueryTreePrinter.printQueryTree(expected, logger);
			TestQueryTreeHelper.equals("", expected, result, logger);

		} catch (MedorException e1) {
			e1.printStackTrace();
			fail(e1.getMessage());
		}
	}
*/

}
