/**
 *  MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */
package org.objectweb.medor.optim.jorm.rdb;

import junit.framework.Test;
import junit.framework.TestSuite;
import junit.textui.TestRunner;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.runtime.ConnectionSpecJDBC;
import org.objectweb.jorm.runtime.MapperJDBC;
import org.objectweb.medor.lib.Log;
import org.objectweb.medor.optim.api.QueryRewriter;
import org.objectweb.medor.optim.jorm.JormFlatten2Rdb;
import org.objectweb.medor.optim.jorm.TestJormGoUpDecodeRule;
import org.objectweb.medor.optim.lib.BasicQueryRewriter;
import org.objectweb.medor.optim.lib.FlattenQueryTreeRule;

import java.util.ArrayList;

/**
 * @author S.Chassande-Barrioz
 */
public class TestRdbJormGoUpDecodeRule extends TestJormGoUpDecodeRule {

	/**
	 * main method to launch the tests manually
	 */
	public static void main(String[] args) {
		TestRunner.run(suite());
	}

	/**
	 * This method creates a TestSuite object with the current tests
	 */
	public static Test suite() {
		return new TestSuite(TestRdbJormGoUpDecodeRule.class);
	}

	public TestRdbJormGoUpDecodeRule() {
		super("TestRdbJormGoUpDecodeRule", Log.MEDOR_PREFIX + ".optim.jorm.rdb.goupdecode");
	}

	public TestRdbJormGoUpDecodeRule(String testName) {
		super(testName, Log.MEDOR_PREFIX + ".optim.jorm.rdb.goupdecode");
	}


    public PMapper newMappper() {
        PMapper m = new MapperJDBC();
        try {
            ConnectionSpecJDBC cs = new ConnectionSpecJDBC("jdbc:product:basename", "java.lang.String");
            m.setConnectionFactory(cs);
        }
        catch (PException e) {
            e.printStackTrace();
        }
        return m;
    }

    public QueryRewriter createQueryRewriter() {
        ArrayList rules = new ArrayList(2);
        rules.add(new FlattenQueryTreeRule());
        rules.add(new JormFlatten2Rdb());
        return new BasicQueryRewriter(rules);
    }
}
